;; Generated package description from tex-parens.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "tex-parens" "0.7.0.20260225.9" "Like lisp.el, but for tex" '((emacs "27.1")) :commit "48b4fcd109e5080fe5e91dccf315d6d3fa270aee" :authors '(("Paul D. Nelson" . "nelson.paul.david@gmail.com")) :maintainer '("Paul D. Nelson" . "nelson.paul.david@gmail.com") :keywords '("tex" "convenience") :url "https://github.com/ultronozm/tex-parens.el")
